/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/* $XConsortium: mfbpushpxl.c,v 1.18 88/09/06 14:53:49 jim Exp $ */

#include "X.h"
#include "gcstruct.h"
#include "screenint.h"
#include "pixmapstr.h"
#include "miscstruct.h"
#include "svga.h"

#define NPT 128

/* mxPushPixels -- squeegees the forground color of pGC through pBitMap
 * into pDrawable.  pBitMap is a stencil (dx by dy of it is used, it may
 * be bigger) which is placed on the drawable at xOrg, yOrg.  Where a 1 bit
 * is set in the bitmap, the fill style is put onto the drawable using
 * the GC's logical function. The drawable is not changed where the bitmap
 * has a zero bit or outside the area covered by the stencil.
 */
void
mxPushPixels(pGC, pBitMap, pDrawable, dx, dy, xOrg, yOrg)
    GCPtr	pGC;
    PixmapPtr	pBitMap;
    DrawablePtr pDrawable;
    int		dx, dy, xOrg, yOrg;
{
    int		h, dxDiv8, ibEnd;
    unsigned char *pwLineStart;
    unsigned char	*pw, *pwEnd;
    unsigned char mask, w;
    int ib;
    int ipt;		/* index into above arrays */
    Bool 	fInBox;
    DDXPointRec	pt[NPT];
    int		width[NPT];

    /* Now scan convert the pixmap and use the result to call fillspans in
     * in the drawable with the original GC */
    ipt = 0;
    dxDiv8 = dx/8;
    for(h = 0; h < dy; h++)
    {

	pw = ((char *)(pBitMap->devPrivate.ptr))+(h * pBitMap->devKind);
	pwLineStart = pw;
	/* Process all words which are fully in the pixmap */
	
	fInBox = FALSE;
	pwEnd = pwLineStart + dxDiv8;
	while(pw  < pwEnd)
	{
	    w = *pw;
	    mask = 0x80;
	    for(ib = 0; ib < 8; ib++)
	    {
		if(w & mask)
		{
		    if(!fInBox)
		    {
			pt[ipt].x = ((pw - pwLineStart) << 3) + ib + xOrg;
			pt[ipt].y = h + yOrg;
			/* start new box */
			fInBox = TRUE;
		    }
		}
		else
		{
		    if(fInBox)
		    {
			width[ipt] = ((pw - pwLineStart) << 3) + 
				     ib + xOrg - pt[ipt].x;
			if (++ipt >= NPT)
			{
			    mxFillSpans(pDrawable, pGC, NPT, pt,
			                      width, SPANS_SORTED, (BoxPtr)0);
			    ipt = 0;
			}
			/* end box */
			fInBox = FALSE;
		    }
		}
		mask >>= 1;
	    }
	    pw++;
	}
	ibEnd = dx & 0x7;
	if(ibEnd)
	{
	    /* Process final partial byte on line */
	    w = *pw;
	    mask = 0x80;
	    for(ib = 0; ib < ibEnd; ib++)
	    {
		if(w & mask)
		{
		    if(!fInBox)
		    {
			/* start new box */
			pt[ipt].x = ((pw - pwLineStart) << 3) + ib + xOrg;
			pt[ipt].y = h + yOrg;
			fInBox = 2;
		    }
		}
		else
		{
		    if(fInBox)
		    {
			/* end box */
			width[ipt] = ((pw - pwLineStart) << 3) + 
				     ib + xOrg - pt[ipt].x;
			if (++ipt >= NPT)
			{
			    mxFillSpans(pDrawable, pGC, NPT, pt,
			                      width, SPANS_SORTED, (BoxPtr)0);
			    ipt = 0;
			}
			fInBox = FALSE;
		    }
		}
		mask >>= 1;
	    }
	}
	/* If scanline ended with last bit set, end the box */
	if(fInBox)
	{
	    width[ipt] = dx + xOrg - pt[ipt].x;
	    if (++ipt >= NPT)
	    {
		mxFillSpans(pDrawable, pGC, NPT, pt, width, SPANS_SORTED, (BoxPtr)0);
		ipt = 0;
	    }
	}
    }
    /* Flush any remaining spans */
    if (ipt)
    {
	mxFillSpans(pDrawable, pGC, ipt, pt, width, SPANS_SORTED, (BoxPtr)0);
    }
}
